using System.Runtime.InteropServices;
using System.Text;
using Sharp.Augeas.Test;
using static Sharp.Augeas.AugeasExtern;

namespace Sharp.Augeas
{
    /// <summary>
    ///  Augeas Core containing the settings of this instance.
    /// </summary>
    public sealed unsafe class Augeas
    {
        #region Flags

        /// <summary>
        /// 
        /// </summary>
        public static int NONE = 0;

        public static int SAVE_BACKUP = (1 << 0);
        public static int SAVE_NEWFILE = (1 << 1);
        public static int TYPE_CHECK = (1 << 2);
        public static int NO_STDINC = (1 << 3);
        public static int SAVE_NOOP = (1 << 4);
        public static int NO_LOAD = (1 << 5);

        public static int NO_MODL_AUTOLOAD = (1 << 6);
        public static int AUG_ENABLE_SPAN = (1 << 7);

        #endregion Flags

        private readonly IntPtr _augeas;

        private HashSet<string> _loadedFiles = new();

        #region Constructor / Destructor

        /// <summary>
        /// Augeas Core Destructor
        /// </summary>
        ~Augeas()
        {
            close_aug(_augeas);
        }

        /// <summary>
        /// Augeas Core constructor
        /// </summary>
        /// <param name="augSettings">Augeas core instance settings</param>
        public Augeas(AugSettings augSettings)
        {
            _augeas = init_aug(augSettings, NO_STDINC | NO_LOAD);
            if (_augeas == IntPtr.Zero)
            {
                throw new InvalidOperationException("Augeas is not a valid instance.");
            }
        }

        /// <summary>
        /// Constructor that reads AUG_LENS_PATH. 
        /// </summary>
        /// <param name="root"></param>
        /// <exception cref="InvalidOperationException">AUG_LENS_PATH environment variable is not set with the augeas lens path.</exception>
        public Augeas(string root)
        {
            var lensPath = Environment.GetEnvironmentVariable("AUG_LENS_PATH");

            if (string.IsNullOrEmpty(lensPath))
            {
                throw new InvalidOperationException(
                    "Can't use root only ctor without a valid AUG_LENS_PATH environment variable.");
            }

            var augSettings = new AugSettings(root, lensPath);

            _augeas = init_aug(augSettings, NO_STDINC | NO_LOAD);

            if (_augeas == IntPtr.Zero)
            {
                throw new InvalidOperationException("Augeas is not a valid instance.");
            }
        }

        #endregion

        #region Augeas Internal Api

        /// <summary>
        /// Prints a preview of the desired segment in <see cref="matchPath"/>
        /// </summary>
        /// <param name="matchPath">Augeas path.</param>
        public void PrintPreview(string matchPath)
        {
            print_preview(_augeas, matchPath);
        }

        /// <summary>
        /// Get the exact value of the matchPath.
        /// Returns a empty string if no match is valid.
        /// </summary>
        /// <param name="matchPath">Augeas path.</param>
        public string GetNode(string matchPath)
        {
            var res = get_node(_augeas, matchPath);
            return res != IntPtr.Zero ? Marshal.PtrToStringAnsi(res) : string.Empty;
        }

        /// <summary>
        /// Binding for aug_load_file 
        /// </summary>
        /// <param name="configurationFilePath">Full path to the configuration file</param>
        /// <returns>True if successfully loaded.</returns>
        public bool LoadFile(string configurationFilePath)
        {
            if (_loadedFiles.Contains(configurationFilePath))
            {
                return true;
            }

            bool success = load_file(_augeas, configurationFilePath);

            _loadedFiles.Add(configurationFilePath);

            return success;
        }

        /// <summary> 
        /// Prints the Aug Tree of a segment 
        /// </summary>
        /// <param name="matchPath"></param>
        /// <param name="filePath"></param>
        private void PrintTree(string matchPath)
        {
            print_tree(_augeas, matchPath);
        }

        private void FreeString(IntPtr str)
        {
            if (str != IntPtr.Zero)
            {
                free_str(str);
            }
        }

        /// <summary>
        /// Returns the tree given the augeas configuration path.
        /// </summary>
        /// <param name="matchPath">Configuration path.</param>
        /// <returns>Dictionary with the Tree</returns>
        public Dictionary<string, string> GetDictionaryTree(string matchPath)
        {
            var result = new Dictionary<string, string>();
            var raw = get_tree(_augeas, matchPath);
            string sb = Marshal.PtrToStringAnsi(raw);
            var lines = sb.Split(";ENDL;");
            foreach (var line in lines)
            {
                var pair = line.Split(";VAL;");
                if (pair.Length == 2)
                {
                    result.Add(pair[0].Remove(0, 3), pair[1]);
                }
            }

            FreeString(raw);
            return result;
        }

        /// <summary>
        /// Sets a value of exactly one node.
        /// </summary>
        /// <param name="matchPath">Configuration path.</param>
        /// <param name="value">The value to set.</param>
        /// <returns>True if the value was set.</returns>
        public bool SetNode(string matchPath, string value)
        {
            bool success = set_node(_augeas, matchPath, value);
            return success;
        }

        /// <summary>
        /// Removes a node from the tree given the <see cref="matchPath"/>
        /// </summary>
        /// <param name="matchPath">Configuration path.</param>
        /// <returns>Removed nodes count.</returns>
        public int RemoveNode(string matchPath)
        {
            int removedCount = rm_node(_augeas, matchPath);
            return removedCount;
        }

        public int InsertNode(string matchPath, string label, int before = 0)
        {
            return insert_node(_augeas, matchPath, label, before);
        }


        /// <summary>
        /// Write all pending changes to disk. Only files that had any changes made to them are written.
        /// If AUG_SAVE_NEWFILE is set in the FLAGS passed to AUG_INIT, create changed files as new files with the extension ".augnew", and leave the
        /// original file unmodified. Otherwise, if AUG_SAVE_BACKUP is set in the FLAGS passed to the constructor,
        /// move the original file to a new file with extension ".augsave".
        /// If neither of these flags is set, overwrite the original file
        /// </summary>
        /// <returns> True on success. Only files that had any changes made to them are written.</returns>
        public bool Save()
        {
            bool saved = save(_augeas);
            return saved;
        }

        /// <summary>
        /// Gets a preview of the configuration segment of <see cref="matchPath"/>.
        /// The file must be loaded first.
        /// </summary>
        /// <param name="matchPath">Augeas path.</param>
        /// <returns></returns>
        public string GetPreview(string matchPath)
        {
            var raw = get_preview(_augeas, matchPath);
            string sb = Marshal.PtrToStringAnsi(raw);
            FreeString(raw);
            return sb;
        }

        #endregion


        /// <summary>
        /// Matches of the path expression PATH in AUG.
        /// </summary>
        /// <param name="matchPath">Augeas path.</param>
        /// <returns></returns>
        public string[] Match(string matchPath)
        {
            sbyte** result = null;
            var count = match(_augeas, matchPath, &result);
            var managedArray = GetStringsAndFreeAllocations(count, result);
            result = null;
            return managedArray;
        }

        private string[] GetStringsAndFreeAllocations(int count, sbyte** list)
        {
            var result = new string[count];

            if (count <= 0) return Array.Empty<string>();

            for (var i = 0; i < count; i++)
            {
                result[i] = Marshal.PtrToStringUTF8((IntPtr)list[i]);
                Marshal.FreeCoTaskMem((IntPtr)list[i]);
            }

            return result;
        }

    }
}