using System.Text;

namespace Sharp.Augeas;

public class Directive : Node
{
    public string Value;
    public List<Argument> Arguments;
    public bool HasArguments => Arguments.Count != 0;

    /// <summary>
    /// Sets a new value for the directive using string literals
    /// The values are separated by spaces, the first value is the directive value,
    /// the other values are values corresponding to the arguments.
    /// </summary>
    /// <param name="newValue"></param>
    public void Set(string newValue)
    {
        var splitted = newValue.Split(" ");
        
        if (splitted.Length > 0)
        {
            SetValue(splitted[0]);
        }

        for (int i = 1; i < splitted.Length; i++)
        {
            var argumentIndex = i - 1;
            Arguments[argumentIndex].SetValue(splitted[i]);
        }
    }
    
    /// <summary>
    /// Set the value of this directive.
    /// Note: This doesn't write any changes to disk. 
    /// </summary>
    /// <param name="newValue">New value.</param>
    /// <returns></returns>
    public bool SetValue(string newValue)
    {
        Value = newValue;
        return _augeas.SetNode(Path, newValue);
    }

    public Directive(Augeas augeas, string id, string path, string value) : base(augeas, path)
    {
        Value = value;
        Id = id;
        Path = path;
        Arguments = new List<Argument>();
    }

    /// <summary>
    /// Add a argument to this node.
    /// </summary>
    /// <param name="argument"></param>
    public void AddArgument(Argument argument)
    {
        argument.Parent = this;
        Arguments.Add(argument);
    }

    public override string ToString()
    {
        var sb = new StringBuilder($"{Value} ");
        
        foreach (var arg in Arguments)
        {
            sb.Append($"{arg} ");
        }
        
        return sb.ToString().TrimEnd();
    }
}